import typing

import yaml

from aok import _definitions
from aok import _operations


class Equals(_definitions.Comparator):
    """Compares two values as an equality."""

    def _compare(
        self,
        observed: typing.Any,
        subset: bool = False,
    ) -> typing.Union[_definitions.Comparison, bool]:
        """Make an equals comparison."""
        return _operations.cast_compatible(self.value, observed) == observed


class Anything(_definitions.Comparator):
    """Allows anything for the given value."""

    def _compare(
        self,
        observed: typing.Any,
        subset: bool = False,
    ) -> typing.Union[_definitions.Comparison, bool]:
        """Anything will always be true."""
        return True

    @classmethod
    def from_yaml(cls, loader: yaml.Loader, node: yaml.Node) -> "Anything":
        return cls(None)


class Less(_definitions.Comparator):
    """Allows anything less than the given value."""

    def _compare(
        self,
        observed: typing.Any,
        subset: bool = False,
    ) -> typing.Union[_definitions.Comparison, bool]:
        """Less than will be true."""
        return _operations.cast_compatible(self.value, observed) > observed


class LessOrEqual(_definitions.Comparator):
    """Allows anything less than or equal the given value."""

    def _compare(
        self,
        observed: typing.Any,
        subset: bool = False,
    ) -> typing.Union[_definitions.Comparison, bool]:
        """Less than or equal will be true."""
        return _operations.cast_compatible(self.value, observed) >= observed


class Greater(_definitions.Comparator):
    """Allows anything greater than the given value."""

    def _compare(
        self,
        observed: typing.Any,
        subset: bool = False,
    ) -> typing.Union[_definitions.Comparison, bool]:
        """Greater than will be true."""
        return _operations.cast_compatible(self.value, observed) < observed


class GreaterOrEqual(_definitions.Comparator):
    """Allows anything greater than or equal to the given value."""

    def _compare(
        self,
        observed: typing.Any,
        subset: bool = False,
    ) -> typing.Union[_definitions.Comparison, bool]:
        """Greater than or equal will be true."""
        return _operations.cast_compatible(self.value, observed) <= observed


class Between(_definitions.Comparator):
    """Allows between the given values."""

    def _compare(
        self,
        observed: typing.Any,
        subset: bool = False,
    ) -> typing.Union[_definitions.Comparison, bool]:
        """Greater than or equal will be true."""
        casted_min = _operations.cast_compatible(self.value["min"], observed)
        casted_max = _operations.cast_compatible(self.value["max"], observed)
        return casted_min <= observed <= casted_max

    @classmethod
    def construct(cls, minimum: typing.Any, maximum: typing.Any) -> "Between":
        return cls({"min": minimum, "max": maximum})

    @classmethod
    def _from_yaml(cls, loader: yaml.Loader, node: yaml.Node) -> "Between":
        if isinstance(node, yaml.SequenceNode):
            loaded = loader.construct_sequence(node, deep=True)
            value = {"min": loaded[0], "max": loaded[1]}
        else:
            value = loader.construct_mapping(node, deep=True)
        return cls(value)


Anything.register()
anything = getattr(Anything, "constructor", Anything)

Between.register()
between = getattr(Between, "constructor", Between)

Equals.register()
equals = getattr(Equals, "constructor", Equals)

Greater.register()
greater = getattr(Greater, "constructor", Greater)

GreaterOrEqual.register()
greater_or_equal = getattr(GreaterOrEqual, "constructor", GreaterOrEqual)

Less.register()
less = getattr(Less, "constructor", Less)

LessOrEqual.register()
less_or_equal = getattr(LessOrEqual, "constructor", LessOrEqual)

__all__ = [
    "Anything",
    "anything",
    "Between",
    "between",
    "Equals",
    "equals",
    "Greater",
    "greater",
    "GreaterOrEqual",
    "greater_or_equal",
    "Less",
    "less",
    "LessOrEqual",
    "less_or_equal",
]



import typing

from aok import _comparisons
from aok import _definitions


def to_comparator(value: typing.Any) -> "_definitions.Comparator":
    """
    Converts the value to a default comparator if not already one. If the value
    is already a comparator, it is returned unchanged.
    """
    if isinstance(value, _definitions.Comparator):
        return value
    elif isinstance(value, dict):
        return _comparisons.Dict(value)

    return _comparisons.Equals(value)


def cast_compatible(
    expectation_value: typing.Any,
    observed_value: typing.Any,
) -> typing.Any:
    """Casts by compatible type, which is limited."""
    ex = expectation_value
    ob = observed_value

    if isinstance(ob, int) and isinstance(ex, str):
        return int(ex)
    elif isinstance(ob, float) and isinstance(ex, str):
        return float(ex)
    elif isinstance(ob, bool) and ex in ("true", "false"):
        return ex == "true"

    return ex



import dataclasses
import typing

import yaml

from aok import _utils


@dataclasses.dataclass()
class Comparator:
    """Okay style comparison class for comparing values."""

    value: typing.Any

    @classmethod
    def operation_name(cls) -> str:
        """..."""
        return _utils.to_snake_case(cls.__name__)

    def _compare(
        self,
        observed: typing.Any,
        subset: bool = False,
    ) -> typing.Union["Comparison", bool]:
        """Make the comparison."""
        return False

    def compare(self, observed: typing.Any, subset: bool = False) -> "Comparison":
        """
        Compares the observed value with the expected value set in this
        comparator object.

        :param observed:
            Value against which to make the comparison.
        :param subset:
            If this is a non-scalar value, e.g. dict, this will allow for dictionary
            subsets to be compared against what is available while accepting other
            values not specified here to be included in the observed value.
        :return:
            Comparator object specifying the result of the comparison with
            supporting data for assertion and other display.
        """
        try:
            result = self._compare(observed, subset)
        except Exception as error:
            return Comparison(
                operation=self.operation_name(),
                success=False,
                expected=self.value,
                observed=observed,
                error=error,
            )

        if isinstance(result, Comparison):
            return result

        return Comparison(
            operation=self.operation_name(),
            success=result,
            expected=self.value,
            observed=observed,
        )

    @classmethod
    def _from_yaml(cls, loader: yaml.Loader, node: yaml.Node) -> "Comparator":
        """Internal yaml node parsing. Defaults to a scalar value."""
        value = loader.construct_scalar(node)
        return cls(value)

    @classmethod
    def parse_yaml(cls, loader: yaml.Loader, node: yaml.Node) -> "Comparator":
        """Yaml alternative constructor that builds the comparator from a yaml node."""
        return cls._from_yaml(loader, node)

    @classmethod
    def register(cls):
        """Registers the comparator with the PyYaml loader."""
        yaml.add_constructor(f"!aok.{cls.operation_name()}", cls.parse_yaml)


@dataclasses.dataclass()
class Comparison:
    """Contains the results and information for an arbitrary comparison."""

    operation: str
    success: bool
    expected: typing.Any
    observed: typing.Any
    children: typing.Dict[typing.Any, "Comparison"] = dataclasses.field(
        default_factory=lambda: {},
    )
    error: typing.Optional[Exception] = None

    def to_diff_info(
        self,
    ) -> typing.Optional[typing.Union[str, typing.Dict[str, typing.Any]]]:
        """Creates a formatted display message for use in assertions."""
        if self.success:
            return None

        if self.children:
            return {
                key: comparison.to_diff_info()
                for key, comparison in self.children.items()
                if not comparison.success
            }

        return "{operation}({expected}, {observed}){error}".format(
            operation=self.operation,
            expected=self.expected,
            observed=self.observed,
            error=f" [ERROR] {self.error}" if self.error else "",
        )

    def failed_keys(self) -> typing.Set[str]:
        """Lists failed absolute keys."""
        if self.success or not self.children:
            return set()

        outputs = []
        for key, child in self.children.items():
            if child.success:
                continue

            if not child.children:
                outputs.append(key)

            outputs += [f"{key}.{child_key}" for child_key in child.failed_keys()]

        return set(outputs)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aok', 'aok._comparisons', 'aok._comparisons.tests']

package_data = \
{'': ['*'], 'aok._comparisons': ['tests/scenarios/*']}

install_requires = \
['PyYAML>=5.3.1,<6.0.0', 'toml>=0.10.2,<0.11.0']

setup_kwargs = {
    'name': 'aok',
    'version': '0.1.1',
    'description': 'Complex dictionary comparisons to simplify testing.',
    'long_description': '# A-OK\n\n[![PyPI version](https://badge.fury.io/py/aok.svg)](https://pypi.org/project/aok/)\n[![build status](https://gitlab.com/rocket-boosters/aok/badges/main/pipeline.svg)](https://gitlab.com/rocket-boosters/aok/commits/main)\n[![coverage report](https://gitlab.com/rocket-boosters/aok/badges/main/coverage.svg)](https://gitlab.com/rocket-boosters/aok/commits/main)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Code style: flake8](https://img.shields.io/badge/code%20style-flake8-white)](https://gitlab.com/pycqa/flake8)\n[![Code style: mypy](https://img.shields.io/badge/code%20style-mypy-white)](http://mypy-lang.org/)\n[![PyPI - License](https://img.shields.io/pypi/l/aok)](https://pypi.org/project/aok/)\n\n*aok* is a library for simplifying the assertions of complex dictionary returns,\nwhich can be used within Python code or loaded via YAML files.\n\n```python\nimport aok\n\nimport my_application\n\n\ndef test_call():\n    """Should return the expected dictionary from my application call."""\n    result: dict = my_application.get_family("Jane Doe")\n    ok = aok.Okay({\n        "mother": {\n            "age": aok.greater_or_equal(50),\n            "full_name": aok.like("* Doe"),\n        },\n        "father": {\n            "age": aok.greater_or_equal(50),\n            "full_name": aok.like("* Doe"),\n        },\n        "younger_brother": {\n            "age": aok.less(10),\n            "full_name": aok.like("* Doe"),\n        }\n    })\n    \n    # Dictionary "result" must be an exact match with the ok expected values.\n    ok.assert_all(result)\n\n    # Dictionary "result" is asserted against ok expected values as a subset, such\n    # that other keys/values may exist within the "result" structure.\n    ok.assert_subset(result)\n```\n\nThe same thing can be archived from a YAML file:\n\n```yaml\nok: !aok\n  mother:\n    age: !aok.greater_or_equal 50\n    full_name: !aok.like \'* Doe\'\n  father:\n    age: !aok.greater_or_equal 50\n    full_name: !aok.like \'* Doe\'\n  younger_brother:\n    age: !aok.less 10\n    full_name: !aok.like \'* Doe\'\n```\n\nand this can be loaded into a test:\n\n```python\nimport aok\nimport yaml\nimport pathlib\n\nimport my_application\n\n\ndef test_call():\n    """Should return the expected dictionary from my application call."""\n    result: dict = my_application.get_family("Jane Doe")\n    data: dict = yaml.full_load(pathlib.Path("expectations.yaml").read_text())\n    ok: aok.Okay = data["ok"]\n    ok.assert_all(result)\n```\n\nThe available comparators are:\n- `aok.anything()` will always succeed, no matter what the observed value is. \n- `aok.between(min, max)` must be greater than or equal to min and less than or equal\n  to the specified min and max values. This can be a numeric or string value.\n- `aok.equals(value)` must be an exact match between the values.\n- `aok.greater(value)` must be greater than the specified value.\n- `aok.greater_or_equal(value)` must be greater than or equal to the specified value.\n- `aok.less(value)` must be less than the specified value.\n- `aok.less_or_equal(value)` must be less than or equal to the specified value.\n- `aok.like(string_value)` string compares against case-insensitive, unix-shell-style\n  wildcard expressions, e.g. "foo*" would match "foo-bar".\n- `aok.like_case(string_value)` string compares against case-sensitive, \n  unix-shell-style wildcard expressions, e.g. "Foo*" would match "Foo-Bar".\n- `aok.match(string_regex_pattern)` matches the string against the specified regex \n  pattern.\n- `aok.not_null(value)` must not be null/None, but can be anything else.\n- `aok.optional(value)` must equal the specified value or be null/None.\n',
    'author': 'Scott Ernst',
    'author_email': 'swernst@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/rocket-boosters/a-ok',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
